{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Harmonic oscillator visualization \n",
    "\n",
    "version 03.0: based on pendulum code; v2 adds driving force curve and changes $\\theta$ to $q$; v3 adds the acceleration plot ($\\ddot q$).\n",
    "\n",
    "* Created 12-Jan-2019 by Dick Furnstahl (furnstahl.1@osu.edu)\n",
    "* Last revised 22-Jan-2021 by Dick Furnstahl (furnstahl.1@osu.edu)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "%matplotlib inline"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "from scipy.integrate import ode, odeint\n",
    "\n",
    "import matplotlib.pyplot as plt"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Harmonic oscillator code"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Harmonic_oscillator():\n",
    "    \"\"\"\n",
    "    Harmonic oscillator class implements the parameters and differential \n",
    "     equation for a damped, driven, simple harmonic oscillator.\n",
    "     \n",
    "    Parameters\n",
    "    ----------\n",
    "    omega0 : float\n",
    "        natural frequency of the oscillator (e.g., \\sqrt{k/m} if a spring) \n",
    "    beta : float\n",
    "        coefficient of damping term (with a factor of 2) \n",
    "    f_ext : float\n",
    "        amplitude of external force (this is f_0 in Taylor) \n",
    "    omega_ext : float\n",
    "        frequency of external force \n",
    "    phi_ext : float\n",
    "        phase angle for external force (taken to be zero in Taylor)\n",
    "\n",
    "    Methods\n",
    "    -------\n",
    "    dy_dt(y, t)\n",
    "        Returns the right side of the differential equation in vector y, \n",
    "        given time t and the corresponding value of y.\n",
    "        \n",
    "    driving_force(t)\n",
    "        Returns the external driving force at time t.\n",
    "    \"\"\"\n",
    "    def __init__(self,\n",
    "                 omega0=1.,\n",
    "                 beta=0.2,\n",
    "                 f_ext=0.2,\n",
    "                 omega_ext=0.689,\n",
    "                 phi_ext=0.\n",
    "                ):\n",
    "        self.omega0 = omega0\n",
    "        self.beta = beta\n",
    "        self.f_ext = f_ext\n",
    "        self.omega_ext = omega_ext\n",
    "        self.phi_ext = phi_ext\n",
    "    \n",
    "    def dy_dt(self, y, t):\n",
    "        \"\"\"\n",
    "        This function returns the right-hand side of the diffeq as a vector: \n",
    "        [d q/dt \n",
    "         d^2q/dt^2]\n",
    "        \n",
    "        Parameters\n",
    "        ----------\n",
    "        y : float\n",
    "            A 2-component vector with y[0] = q(t) and y[1] = dq/dt\n",
    "        t : float\n",
    "            time \n",
    "                    \n",
    "        \"\"\"\n",
    "        F_ext = self.driving_force(t)\n",
    "        return [y[1], -self.omega0**2*y[0] - 2.*self.beta*y[1] + F_ext]\n",
    "    \n",
    "    def driving_force(self, t):\n",
    "        \"\"\"\n",
    "        This function returns the value of the driving force at time t.\n",
    "        \"\"\"\n",
    "        return self.f_ext * np.cos(self.omega_ext*t + self.phi_ext)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "def plot_y_vs_x(x, y, axis_labels=None, label=None, title=None, \n",
    "                color=None, linestyle=None, ax=None):\n",
    "    \"\"\"\n",
    "    Generic plotting function: return a figure axis with a plot of y vs. x,\n",
    "    with line color and style, title, axis labels, and line label\n",
    "    \"\"\"\n",
    "    if ax is None:        # if the axis object doesn't exist, make one\n",
    "        ax = plt.gca()\n",
    "\n",
    "    line, = ax.plot(x, y, label=label, color=color, linestyle=linestyle)\n",
    "    if label is not None:    # if a label if passed, show the legend\n",
    "        ax.legend()\n",
    "    if title is not None:    # set a title if one if passed\n",
    "        ax.set_title(title)\n",
    "    if axis_labels is not None:  # set x-axis and y-axis labels if passed  \n",
    "        ax.set_xlabel(axis_labels[0])\n",
    "        ax.set_ylabel(axis_labels[1])\n",
    "\n",
    "    return ax, line\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Interface using ipywidgets with interactive_output\n",
    "\n",
    "We'll make a more elaborate interface so we can adjust all of the parameters."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Import the widgets we will use (add more if needed!) \n",
    "import ipywidgets as widgets\n",
    "from ipywidgets import HBox, VBox, Layout, Tab, Label, Checkbox\n",
    "from ipywidgets import FloatSlider, IntSlider, Play, Dropdown, HTMLMath \n",
    "\n",
    "from IPython.display import display\n",
    "from time import sleep"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "# This function generates the main output here, which is a grid of plots\n",
    "def ho_plots(q_vs_time_plot=True, q_dot_vs_time_plot=True,\n",
    "             q_ddot_vs_time_plot=True,\n",
    "             phase_space_plot=True, driving_curve=True,\n",
    "             omega0=10.*np.pi, beta=np.pi/2., \n",
    "             f_ext=1000., omega_ext=2.*np.pi, phi_ext=0., \n",
    "             q0=0.0, q_dot0=0.0,  \n",
    "             t_start=0, t_end=10, delta_t=0.01, plot_start=0,\n",
    "             font_size=18):\n",
    "    \"\"\"\n",
    "    Create plots for interactive_output according to the inputs.\n",
    "    \n",
    "    Based on generating a Harmonic_oscillator instance and associated graphs.\n",
    "    \n",
    "    Notes\n",
    "    -----\n",
    "        1. We generate a new Harmonic_oscillator instance every time *and* \n",
    "            solve the ODE every time, even if the only change is to parameters\n",
    "            like t_start and t_end.  Should we care or is this just so\n",
    "            cheap to recalculate that it doesn't matter?\n",
    "            How could we structure this differently?\n",
    "        2. Should we delete ho1 at some point?  E.g., is there a memory issue?\n",
    "    \"\"\"\n",
    "    \n",
    "    # add delta_t it goes at least to t_end (probably should use linspace!)\n",
    "    t_pts = np.arange(t_start, t_end+delta_t, delta_t)  \n",
    "        \n",
    "    # Instantiate an oscillator with the passed (or default) values of the \n",
    "    #  natural frequency omega0, damping beta, driving amplitude, frequency, \n",
    "    #  and phase (f_ext, omega_ext, phi_ext).\n",
    "    ho1 = Harmonic_oscillator(omega0=omega0, beta=beta, f_ext=f_ext, \n",
    "                              omega_ext=omega_ext, phi_ext=phi_ext)\n",
    "    y0 = [q0, q_dot0]  # initial conditions for the oscillator ODE\n",
    "\n",
    "    # ODE solver parameters\n",
    "    abserr = 1.0e-8\n",
    "    relerr = 1.0e-6\n",
    "\n",
    "    # For now we solve with odeint; give more options in the future.\n",
    "    #  The .T is for transpose, so that the matrix from odeint can changed\n",
    "    #  to the correct form for reading off q and q_dot.\n",
    "    q, q_dot = odeint(ho1.dy_dt, y0, t_pts,\n",
    "                              atol=abserr, rtol=relerr).T\n",
    "    q_ddot = np.gradient(q_dot, t_pts)\n",
    "    # also calculate the driving force for the same t_pts\n",
    "    driving = ho1.driving_force(t_pts)\n",
    "    \n",
    "    # Update the common font size\n",
    "    plt.rcParams.update({'font.size': font_size})\n",
    " \n",
    "    # Labels for individual plot axes\n",
    "    q_vs_time_labels = (r'$t$', r'$q$')\n",
    "    q_dot_vs_time_labels = (r'$t$', r'$dq/dt$')\n",
    "    q_ddot_vs_time_labels = (r'$t$', r'$d^2q/dt^2$')\n",
    "    phase_space_labels = (r'$q$', r'$dq/dt$')\n",
    "    \n",
    "    # Figure out how many rows and columns [one row for now]\n",
    "    plot_flags = [q_vs_time_plot, q_dot_vs_time_plot, q_ddot_vs_time_plot, phase_space_plot]\n",
    "    plot_num = plot_flags.count(True)\n",
    "    plot_rows = 1\n",
    "    figsize_rows = plot_rows*6\n",
    "    plot_cols = plot_num\n",
    "    figsize_cols = min(plot_cols*8, 16)  # at most 16\n",
    "    \n",
    "    # Make the plot!\n",
    "    fig = plt.figure(figsize=(figsize_cols,figsize_rows))\n",
    "    #, axes = plt.subplots(plot_rows, plot_cols, \n",
    "                             \n",
    "    # finds nearest index to plot_start in t_pts array                      \n",
    "    start_index = (np.fabs(t_pts-plot_start)).argmin() \n",
    "    \n",
    "    next_axis = 1  # keep track of the axis number\n",
    "    if q_vs_time_plot:\n",
    "        ax_q = fig.add_subplot(plot_rows, plot_cols, next_axis)                  \n",
    "        plot_y_vs_x(t_pts, q, axis_labels=q_vs_time_labels, \n",
    "                    label='oscillator', title=r'$q$ vs. time', \n",
    "                    ax=ax_q)    \n",
    "        # add a line where the phase space plot starts\n",
    "        ax_q.axvline(t_pts[start_index], lw=3, color='red')\n",
    "                          \n",
    "        if driving_curve:\n",
    "           ax_driving = ax_q.twinx()\n",
    "           plot_y_vs_x(t_pts, driving, ax=ax_driving, color='red',\n",
    "                       linestyle='dotted', label='driving')      # add 'driving label?'                   \n",
    "        next_axis += 1\n",
    "    \n",
    "    if q_dot_vs_time_plot:\n",
    "        ax_q_dot = fig.add_subplot(plot_rows, plot_cols, next_axis)                  \n",
    "        plot_y_vs_x(t_pts, q_dot, axis_labels=q_dot_vs_time_labels, \n",
    "                    label='oscillator', title=r'$dq/dt$ vs. time', \n",
    "                    ax=ax_q_dot)    \n",
    "        # add a line where the phase space plot starts\n",
    "        ax_q_dot.axvline(t_pts[start_index], lw=3, color='red')\n",
    "                          \n",
    "        if driving_curve:\n",
    "           ax_driving2 = ax_q_dot.twinx()\n",
    "           plot_y_vs_x(t_pts, driving, ax=ax_driving2, color='red',\n",
    "                       linestyle='dotted')  # add 'driving label?'                        \n",
    "        next_axis += 1\n",
    "    \n",
    "    if q_ddot_vs_time_plot:\n",
    "        ax_q_ddot = fig.add_subplot(plot_rows, plot_cols, next_axis)                  \n",
    "        plot_y_vs_x(t_pts, q_ddot, axis_labels=q_ddot_vs_time_labels, \n",
    "                    label='oscillator', title=r'$d^2q/dt^2$ vs. time', \n",
    "                    ax=ax_q_ddot)    \n",
    "        # add a line where the phase space plot starts\n",
    "        ax_q_ddot.axvline(t_pts[start_index], lw=3, color='red')\n",
    "                          \n",
    "        if driving_curve:\n",
    "           ax_driving3 = ax_q_ddot.twinx()\n",
    "           plot_y_vs_x(t_pts, driving, ax=ax_driving3, color='red',\n",
    "                       linestyle='dotted')  # add 'driving label?'                        \n",
    "        next_axis += 1\n",
    "\n",
    "    if phase_space_plot:\n",
    "        ax_phase_space = fig.add_subplot(plot_rows, plot_cols, next_axis)                  \n",
    "        plot_y_vs_x(q[start_index:-1], q_dot[start_index:-1], \n",
    "                    axis_labels=phase_space_labels, title='State space', \n",
    "                    ax=ax_phase_space)    \n",
    "        next_axis += 1\n",
    "    \n",
    "    fig.tight_layout()\n",
    "    \n",
    "    return fig\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {
    "scrolled": false
   },
   "outputs": [
    {
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "model_id": "493512a8af8643e4b53e9a3d1e02f5e9",
       "version_major": 2,
       "version_minor": 0
      },
      "text/plain": [
       "VBox(children=(Tab(children=(VBox(children=(HBox(children=(FloatSlider(value=31.41592653589793, continuous_upd…"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "# Widgets for the various inputs.\n",
    "#   For any widget, we can set continuous_update=False if we don't want the \n",
    "#    plots to shift until the selection is finished (particularly relevant for \n",
    "#    sliders).\n",
    "\n",
    "# Widgets for the plot choice (plus a label out front)\n",
    "plot_choice_w = Label(value='Which plots: ',layout=Layout(width='100px'))\n",
    "def plot_choice_widget(on=True, plot_description=None):\n",
    "    \"\"\"Makes a Checkbox to select whether to show a plot.\"\"\"\n",
    "    return Checkbox(value=on, description=plot_description,\n",
    "                  disabled=False, indent=False, layout=Layout(width='150px'))\n",
    "q_vs_time_plot_w = plot_choice_widget(True, r'$q$ vs. time')\n",
    "q_dot_vs_time_plot_w = plot_choice_widget(False, r'$dq/dt$ vs. time')\n",
    "q_ddot_vs_time_plot_w = plot_choice_widget(False, r'$d^2q/dt^2$ vs. time')\n",
    "phase_space_plot_w = plot_choice_widget(True, 'state space')\n",
    "driving_curve_w = plot_choice_widget(False, 'driving force')\n",
    "\n",
    "\n",
    "# Widgets for the oscillator parameters (all use FloatSlider, so we made \n",
    "#  it a function)\n",
    "def float_widget(value, min, max, step, description, format):\n",
    "    \"\"\"Makes a FloatSlider with the passed parameters and continuous_update\n",
    "       set to False.\"\"\"\n",
    "    slider_border = Layout(border='solid 1.0px')\n",
    "    return FloatSlider(value=value,min=min,max=max,step=step,disabled=False,\n",
    "                       description=description,continuous_update=False,\n",
    "                       orientation='horizontal',layout=slider_border,\n",
    "                       readout=True,readout_format=format)\n",
    "\n",
    "omega0_w = float_widget(value=10.*np.pi, min=0.0, max=20.*np.pi, step=0.1,\n",
    "                        description=r'natural $\\omega_0$:', format='.2f')\n",
    "beta_w = float_widget(value=np.pi/2., min=0.0, max=2.*np.pi, step=0.1,\n",
    "                       description=r'damping $\\beta$:', format='.2f')\n",
    "f_ext_w = float_widget(value=1000., min=0.0, max=2000., step=1.,\n",
    "                       description=r'drive $f_{\\rm ext}$:', format='.0f')\n",
    "omega_ext_w = float_widget(value=2.*np.pi, min=0.0, max=6.*np.pi, step=0.1,\n",
    "                       description=r'freq. $\\omega_{\\rm ext}$:', format='.2f')\n",
    "phi_ext_w = float_widget(value=0.0, min=0, max=2.*np.pi, step=0.1,\n",
    "                         description=r'phase $\\phi_{\\rm ext}$:', format='.1f')\n",
    "\n",
    "# Widgets for the initial conditions\n",
    "q0_w = float_widget(value=0.0, min=-2.*np.pi, max=2.*np.pi, step=0.1,\n",
    "                        description=r'$q_0$:', format='.1f')\n",
    "q_dot0_w = float_widget(value=0.0, min=-100., max=100., step=1.,\n",
    "                            description=r'$(dq/dt)_0$:', format='.1f')\n",
    "\n",
    "# Widgets for the plotting parameters\n",
    "t_start_w = float_widget(value=0., min=0., max=10., step=1.,\n",
    "                         description='t start:', format='.1f') \n",
    "t_end_w = float_widget(value=5., min=0., max=20., step=1.,\n",
    "                       description='t end:', format='.1f')\n",
    "delta_t_w = float_widget(value=0.001, min=0.001, max=0.1, step=0.001,\n",
    "                         description='delta t:', format='.3f')\n",
    "plot_start_w = float_widget(value=0., min=0., max=20., step=1.,\n",
    "                            description='start plotting:', format='.1f')\n",
    "\n",
    "# Widgets for the styling parameters\n",
    "font_size_w = Dropdown(options=['12', '16', '18', '20', '24'], value='18',\n",
    "                       description='Font size:',disabled=False,\n",
    "                       continuous_update=False,layout=Layout(width='140px'))\n",
    "\n",
    "\n",
    "############## Begin: Explicit callback functions #######################\n",
    "\n",
    "# Make sure that t_end is at least t_start + 10\n",
    "def update_t_end(*args):\n",
    "    if t_end_w.value < t_start_w.value:\n",
    "        t_end_w.value = t_start_w.value + 10     \n",
    "t_end_w.observe(update_t_end, 'value')\n",
    "t_start_w.observe(update_t_end, 'value')\n",
    "\n",
    "\n",
    "# Make sure that plot_start is at least t_start and less than t_end\n",
    "def update_plot_start(*args):\n",
    "    if plot_start_w.value < t_start_w.value:\n",
    "        plot_start_w.value = t_start_w.value\n",
    "    if plot_start_w.value > t_end_w.value:\n",
    "        plot_start_w.value = t_end_w.value\n",
    "plot_start_w.observe(update_plot_start, 'value')\n",
    "t_start_w.observe(update_plot_start, 'value')\n",
    "t_end_w.observe(update_plot_start, 'value')\n",
    "\n",
    "\n",
    "############## End: Explicit callback functions #######################\n",
    "\n",
    "# Set up the interactive_output widget \n",
    "plot_out = widgets.interactive_output(ho_plots,\n",
    "                          dict(\n",
    "                          q_vs_time_plot=q_vs_time_plot_w,\n",
    "                          q_dot_vs_time_plot=q_dot_vs_time_plot_w,\n",
    "                          q_ddot_vs_time_plot=q_ddot_vs_time_plot_w,\n",
    "                          phase_space_plot=phase_space_plot_w,\n",
    "                          driving_curve = driving_curve_w,\n",
    "                          omega0=omega0_w,\n",
    "                          beta=beta_w,\n",
    "                          f_ext=f_ext_w,\n",
    "                          omega_ext=omega_ext_w,\n",
    "                          phi_ext=phi_ext_w,\n",
    "                          q0=q0_w,\n",
    "                          q_dot0=q_dot0_w,\n",
    "                          t_start=t_start_w,\n",
    "                          t_end=t_end_w, \n",
    "                          delta_t=delta_t_w,    \n",
    "                          plot_start=plot_start_w, \n",
    "                          font_size=font_size_w)\n",
    "                       )\n",
    "\n",
    "# Now do some manual layout, where we can put the plot anywhere using plot_out\n",
    "hbox1 = HBox([plot_choice_w, q_vs_time_plot_w, q_dot_vs_time_plot_w, q_ddot_vs_time_plot_w,\n",
    "              phase_space_plot_w, driving_curve_w]) #  choice of plots to show\n",
    "hbox2 = HBox([omega0_w, f_ext_w, omega_ext_w, phi_ext_w]) # external driving parameters\n",
    "hbox3 = HBox([q0_w, q_dot0_w, beta_w]) # initial conditions and damping\n",
    "hbox4 = HBox([t_start_w, t_end_w, delta_t_w, plot_start_w]) # time, plot ranges\n",
    "hbox5 = HBox([font_size_w]) # font size\n",
    "\n",
    "# We'll set up Tabs to organize the controls.  The Tab contents are declared\n",
    "#  as tab0, tab1, ... (probably should make this a list?) and the overall Tab\n",
    "#  is called tab (so its children are tab0, tab1, ...).\n",
    "tab_height = '40px'  # Fixed minimum height for all tabs. Specify another way?\n",
    "tab0 = VBox([hbox2, hbox3], layout=Layout(min_height=tab_height))\n",
    "tab1 = VBox([hbox1, hbox4], layout=Layout(min_height=tab_height))\n",
    "tab2 = VBox([hbox5], layout=Layout(min_height=tab_height))\n",
    "\n",
    "tab = Tab(children=[tab0, tab1, tab2])\n",
    "tab.set_title(0, 'Physics')\n",
    "tab.set_title(1, 'Plotting')\n",
    "tab.set_title(2, 'Styling')\n",
    "\n",
    "# Release the Kraken!\n",
    "vbox2 = VBox([tab, plot_out])\n",
    "display(vbox2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
